{{ header }}
    - dimension: {{ elem.dim }}
    - center: {{ elem.center }}
    - v1: {{ elem.v1 }}
    - v2: {{ elem.v2 }}
    - v3: {{ elem.v3 }}
    - label: {{ elem.label }}
    - type: {{ type }}